#include <stdlib.h>
#include "tri_lisse.h"

liste* creer_liste_vide() {
    liste* l = (liste*) malloc(sizeof(liste));
    l->premier_noeud = NULL;
    return l;
}

void libere_liste( liste* l ) {
    noeud_liste* noeud;
    while( l->premier_noeud ) {
        noeud = l->premier_noeud->noeud_suivant;
        free(l->premier_noeud);
        l->premier_noeud = noeud;
    }
    free(l);
}

// Prérequis : i est un indice valide de la liste
int nth( liste* l, int i ) {
    noeud_liste* noeud_courant = l->premier_noeud;

    for (; i > 0; --i)
        noeud_courant = noeud_courant->noeud_suivant;
    return  noeud_courant->valeur;
}

void insere( liste* l, int x ) {
    noeud_liste *noeud_courant = l->premier_noeud, *nouveau_noeud;
    noeud_liste** p_precedent = &(l->premier_noeud);

    nouveau_noeud = (noeud_liste*) malloc(sizeof(noeud_liste));
    nouveau_noeud->valeur = x;
    for (noeud_courant = l->premier_noeud ;
         noeud_courant && noeud_courant->valeur <= x;
         noeud_courant = noeud_courant->noeud_suivant) {
        p_precedent = &(noeud_courant->noeud_suivant);
    }
        nouveau_noeud->noeud_suivant = noeud_courant;
        *p_precedent = nouveau_noeud;
}

liste* tri_insertion( liste* l ) {
    noeud_liste* noeud_courant;
    liste* l_triee = creer_liste_vide();

    for (noeud_courant = l->premier_noeud ;
         noeud_courant ;
         noeud_courant = noeud_courant->noeud_suivant)
        insere(l_triee, noeud_courant->valeur);

    return l_triee;
}

liste* e( int u0, int n ) {
    int u = u0;
    liste* e = creer_liste_vide();
    insere(e,u);
    for (int i = 0 ; i < n ; ++i) {
        u = (15091 * u) % 64007;
        insere(e,u);
    }
    return e;
}

arbre* creer_arbre( int x, arbre* fg, arbre* fd ) {
    noeud_arbre* racine = (noeud_arbre*) malloc(sizeof(noeud_arbre));
    arbre* a = (arbre*) malloc(sizeof(arbre));

    racine->valeur = x;
    racine->fils_gauche = fg ? fg->racine : NULL;
    free(fg);
    racine->fils_droit = fd ? fd->racine : NULL;
    free(fd);
    a->racine = racine;
    return a;
}

void libere_noeud_arbre( noeud_arbre* a ) {
    if(!a)
        return;
    libere_noeud_arbre(a->fils_gauche);
    libere_noeud_arbre(a->fils_droit);
    free(a);
}

void libere_arbre( arbre* a ) {
    libere_noeud_arbre(a->racine);
    free(a);
}

// Prérequis : a non vide
int min_tas( arbre* a ) {
    return a->racine->valeur;
}

// Prérequis : a non vide
int min_quasi( arbre* a ) {
    int min = a->racine->valeur;
    if (!a->racine->fils_gauche)
        return min;
    if ( a->racine->fils_gauche->valeur < min )
        min = a->racine->fils_gauche->valeur;
    if ( a->racine->fils_droit->valeur < min )
        min = a->racine->fils_droit->valeur;
    return min;
}

void percole( arbre* a ) {
    if (!a->racine )
        return;

    int min;
    arbre tmp_a = { NULL };
    noeud_arbre* noeud_courant = a->racine ;
    noeud_arbre* noeud_suivant;

    while( noeud_courant->fils_gauche ) {
        tmp_a.racine = noeud_courant;
        min = min_quasi(&tmp_a);
        if (min == noeud_courant->valeur)
            break;

        if (noeud_courant->fils_gauche->valeur == min)
            noeud_suivant = noeud_courant->fils_gauche;
        else
            noeud_suivant = noeud_courant->fils_droit;

        noeud_suivant->valeur = noeud_courant->valeur;
        noeud_courant->valeur = min;
        noeud_courant = noeud_suivant;
    }
}

liste_tas* creer_liste_tas() {
    liste_tas* h = (liste_tas*) malloc(sizeof(liste_tas));
    h->premier_noeud = NULL;
    h->dernier_noeud = NULL;
    return h;
}

void libere_liste_tas( liste_tas* h ) {
    noeud_liste_tas* noeud;
    while( h->premier_noeud ) {
        noeud = h->premier_noeud->noeud_suivant;
        libere_arbre(h->premier_noeud->tas);
        free(h->premier_noeud);
        h->premier_noeud = noeud;
    }
    free(h);
}

void insere_tas(liste_tas* h, arbre* a, int x ) {
    noeud_liste_tas* noeud =
        (noeud_liste_tas*) malloc(sizeof(noeud_liste_tas));
    noeud->taille = x;
    noeud->tas = a;
    noeud->noeud_suivant = h->premier_noeud;
    noeud->noeud_precedent = NULL;
    if(h->premier_noeud)
        h->premier_noeud->noeud_precedent = noeud;
    else
        h->dernier_noeud = noeud;
    h->premier_noeud = noeud;
}

// Prérequis h non vide
void extrait_tas(liste_tas* h, arbre** a, int* x) {
    noeud_liste_tas *n1 = h->premier_noeud;
    *a = n1->tas;
    *x = n1->taille;
    h->premier_noeud = n1->noeud_suivant;
    if(h->premier_noeud)
        h->premier_noeud->noeud_precedent = NULL;
    else
        h->dernier_noeud = NULL;
    free(n1);
}

void ajoute( liste_tas* h, int x ) {
    arbre *ax, *a1, *a2;
    int t1, t2;
    if(!h->premier_noeud
       || !h->premier_noeud->noeud_suivant
       || h->premier_noeud->taille != h->premier_noeud->noeud_suivant->taille) {
        ax = creer_arbre(x,NULL,NULL);
        insere_tas(h, ax, 1);
    } else {
        extrait_tas(h, &a1, &t1);
        extrait_tas(h, &a2, &t2);
        ax = creer_arbre(x,a1,a2);
        insere_tas(h, ax, t1 + t2 + 1);
        percole(h->premier_noeud->tas);
    }
}

liste_tas* constr_liste_tas( liste* l ) {
    liste_tas* h = creer_liste_tas();

    for (noeud_liste* noeud_courant = l->premier_noeud ;
         noeud_courant ;
         noeud_courant = noeud_courant->noeud_suivant )
        ajoute(h,noeud_courant->valeur);
    return h;
}

//Prerequis : a et b sont non vides
void echange_racine( arbre* a, arbre* b ) {
    a->racine->valeur ^= b->racine->valeur;
    b->racine->valeur ^= a->racine->valeur;
    a->racine->valeur ^= b->racine->valeur;
}

// Prérequis : h vérifie RO et t est la taille de a
void insere_tas_ro ( liste_tas* h, arbre* a, int t ) {
    noeud_liste_tas *noeud_suivant, *noeud_courant;

    insere_tas(h, a, t);

    noeud_courant = h->premier_noeud;
    noeud_suivant = noeud_courant->noeud_suivant;
    while (noeud_suivant
        && min_quasi(noeud_courant->tas) > noeud_suivant->tas->racine->valeur )  {
        echange_racine(noeud_courant->tas, noeud_suivant->tas);
        noeud_courant = noeud_suivant;
        noeud_suivant = noeud_courant->noeud_suivant;
    }
    percole(noeud_courant->tas);
}

// Prérequis : h vérifie TC
void tri_racine( liste_tas** h ) {
    liste_tas* h_triee = creer_liste_tas();

    for (noeud_liste_tas* noeud_courant = (*h)->dernier_noeud ;
         noeud_courant ;
         noeud_courant = noeud_courant->noeud_precedent )
        insere_tas_ro(h_triee, noeud_courant->tas, noeud_courant->taille);
   free(*h);
   *h = h_triee;
}

int extraire( liste_tas* h) {
    arbre *a, *a1, *a2;
    int x, t;

    extrait_tas(h,&a,&t);
    x = a->racine->valeur;
    if(a->racine->fils_gauche) {
        a1 = (arbre*) malloc(sizeof(arbre));
        a1->racine = a->racine->fils_gauche ;
        a2 = (arbre*) malloc(sizeof(arbre));
        a2->racine = a->racine->fils_droit ;
        t = ( t - 1 ) / 2;
        insere_tas_ro(h, a2, t );
        insere_tas_ro(h, a1, t );
    }

    free(a->racine);
    free(a);
    return x;
}

liste* tri_lisse( liste* l) {
    liste* l_triee = creer_liste_vide();
    liste_tas* h = constr_liste_tas(l);
    tri_racine(&h);
    while(h->premier_noeud)
        insere(l_triee, extraire(h));
    free(h);
    return l_triee;
}
